/*********************************************************************************************************************
  Seja S um conjunto de segmentos horizontais. Seja s_v um segmento vertical. 
  Queremos encontrar os segmentos de $S$ que atravessam s_v.
*********************************************************************************************************************/

#include <Arvore_Prioridade/arvore_prioridade_x_min.h>
#include <Arvore_Prioridade/arvore_prioridade_x_max.h>

struct noArvoreIntervalosHIndireto {
  int p_med;
  noArvorePrioridadexmSIndireto * arvorexmS;
  noArvorePrioridadexMSIndireto * arvorexMS;
  struct noArvoreIntervalosHIndireto * esq;
  struct noArvoreIntervalosHIndireto * dir;
};

class ArvoreIntervalosHIndireto{
 public:
  /*
    Recebe: Um vetor de segmentos S ordenado pela x-coordenada dos pontos extremos esquerdos.
    Devolve: Uma arvore de intervalos.
  */
  static noArvoreIntervalosHIndireto * ConstroiArvoreIntervalosHIndireto (Segmento * vetorGlobal, int * S, int n){
    noArvoreIntervalosHIndireto * v = NULL;
    char * ponto_e = "pe";
    char * ponto_d = "pd";
    if (n > 0){
      int * P_ex = (int *) malloc (sizeof(int) * n);
      int * P_ey = (int *) malloc (sizeof(int) * n);
      int * P_dx = (int *) malloc (sizeof(int) * n);
      int * P_dy = (int *) malloc (sizeof(int) * n);
      int * S_e;
      int * S_d;
      
      int i, n_e, n_d, n_med, meio;

      v = (noArvoreIntervalosHIndireto *) malloc (sizeof(noArvoreIntervalosHIndireto));

      meio = (int) (n / 2.0 + 0.5);
      
      S_e = (int *) malloc (sizeof(int) * meio);
      S_d = (int *) malloc (sizeof(int) * meio);
      
      v->p_med = S[meio-1]; /*Isso eh um segmento mas tenho que considerar o seu ponto extremo esquerdo.*/

      i = 0, n_e = 0, n_med = 0;

      while (i < n && vetorGlobal[S[i]].get_pe().getX() <= vetorGlobal[v->p_med].get_pe().getX()){
	    if (vetorGlobal[S[i]].get_pd().getX() < vetorGlobal[v->p_med].get_pe().getX()){
	      S_e[n_e] = S[i];
	      n_e++;
	    }
	    else{
	      P_dx[n_med] = S[i];
	      P_dy[n_med] = S[i];
	      P_ex[n_med] = S[i];/*P_ex já está ordenado.*/
	      P_ey[n_med] = S[i];
	      n_med++;
	    }
	    i++;
      }

      n_d = 0;

      while (i < n){
	    S_d[n_d] = S[i];
	    n_d++;
	    i++;
      }

      /*Lembre que P_ex ja esta ordenado. Isso vale porque S está ordenado pela x-coordenada dos pontos extremos esq.*/
      sort(P_ey, P_ey + n_med, compare_pe_y);
      /*Lembre que S_mpe tmb ja esta ordenado pelo msm motivo.*/
           
      v->arvorexmS = ArvorePrioridadexmSIndireto::ConstroiArvorePrioridadexmS(vetorGlobal, P_ex, P_ey, n_med, ponto_e);

      sort(P_dx, P_dx + n_med, compare_pd_x);
      sort(P_dy, P_dy + n_med, compare_pd_y);
      
      v->arvorexMS = ArvorePrioridadexMSIndireto::ConstroiArvorePrioridadexMS(vetorGlobal, P_dx, P_dy, n_med, ponto_d);

      v->esq = ConstroiArvoreIntervalosHIndireto(vetorGlobal, S_e, n_e);
      v->dir = ConstroiArvoreIntervalosHIndireto(vetorGlobal, S_d, n_d);
      free(P_ex);
      free(P_ey);
      free(P_dx);
      free(P_dy);
      free(S_e);
      free(S_d);      
    }
    return v;
  }
  
  static void ConsultaArvoreIntervalosHIndireto (Segmento * vetorGlobal, noArvoreIntervalosHIndireto * v, Segmento Seg, list <int> &l){
    
    Janela W_ilim;
    if (v != NULL){
      //if (Seg.get_pe().xMaior(v->p_med)){
      if (Seg.get_pe().getX() > vetorGlobal[v->p_med].get_pe().getX()){
	    W_ilim = Janela(Seg.get_pe(), Ponto(MaxInteger, Seg.get_pd().getY()));
	    ArvorePrioridadexMSIndireto::ConsultaArvorePrioridadexMS(vetorGlobal, *(v->arvorexMS), W_ilim, l);
	    ConsultaArvoreIntervalosHIndireto (vetorGlobal, v->dir, Seg, l);
      }
      else{
	    W_ilim = Janela(Ponto(MinInteger, Seg.get_pe().getY()), Ponto(Seg.get_pd().getX(), Seg.get_pd().getY()));
	    ArvorePrioridadexmSIndireto::ConsultaArvorePrioridadexmS(vetorGlobal, *(v->arvorexmS), W_ilim, l);
	    ConsultaArvoreIntervalosHIndireto (vetorGlobal, v->esq, Seg, l);
      }
    }
  }

  static void ImprimeArvoreIntervalosH (Segmento * vetorGlobal, noArvoreIntervalosHIndireto * v){
    if (v != NULL){
      printf("\n");
      Ponto::ImprimePonto(vetorGlobal[v->p_med].get_pe());
      printf("\nW-:\n");
      ArvorePrioridadexmSIndireto::ImprimeArvorePrioridadexmS(vetorGlobal, *(v->arvorexmS));
      printf("\nW+:\n");
      ArvorePrioridadexMSIndireto::ImprimeArvorePrioridadexMS(vetorGlobal, *(v->arvorexMS));
      ImprimeArvoreIntervalosH (vetorGlobal, v->esq);
      ImprimeArvoreIntervalosH (vetorGlobal, v->dir);
    }
  }

  static void inicializa (noArvoreIntervalosHIndireto * T){
    noArvoreIntervalosHIndireto * v = T;
    if (v != NULL){
      ArvorePrioridadexmSIndireto::inicializa(v->arvorexmS);
      ArvorePrioridadexMSIndireto::inicializa(v->arvorexMS);
      ArvoreIntervalosHIndireto::inicializa(v->esq);
      ArvoreIntervalosHIndireto::inicializa(v->dir);
      free(v);
    }
  }

  static size_t consumo_espaco (noArvoreIntervalosHIndireto * T){
    size_t c = 0;
    noArvoreIntervalosHIndireto * v = T;
    if (v != NULL){
      c = sizeof(v) + sizeof(v->p_med);
      c += ArvorePrioridadexmSIndireto::consumo_espaco(v->arvorexmS);
      c += ArvorePrioridadexMSIndireto::consumo_espaco(v->arvorexMS);
      c += ArvoreIntervalosHIndireto::consumo_espaco(v->esq);
      c += ArvoreIntervalosHIndireto::consumo_espaco(v->dir);
    }
    return c;
  }
};
