/************************************************************************************************************************
  Seja S um conjunto de segmentos verticais. Seja s_h um segmento horizontais. 
  Queremos encontrar os segmentos de $S$ que atravessam s_h.
*************************************************************************************************************************/

#include <Arvore_Prioridade/arvore_prioridade_y_min.h>
#include <Arvore_Prioridade/arvore_prioridade_y_max.h>

struct noArvoreIntervalosVIndireto {
  int p_med;
  noArvorePrioridadeymSIndireto * arvoreymS;
  noArvorePrioridadeyMSIndireto * arvoreyMS;
  struct noArvoreIntervalosVIndireto * abaixo;
  struct noArvoreIntervalosVIndireto * acima;
};

class ArvoreIntervalosVIndireto{
 public:
  /*
    Recebe: Um vetor de segmentos S ordenado pela y-coordenada dos pontos extremos esquerdos.
    Devolve: Uma arvore de intervalos.
  */
  static noArvoreIntervalosVIndireto * ConstroiArvoreIntervalosVIndireto (Segmento * vetorGlobal, int * S, int n){
    noArvoreIntervalosVIndireto * v = NULL;
    if (n > 0){
      char * ponto_e = "pe";
      char * ponto_d = "pd";
      int * P_abaixo_x = (int *) malloc (sizeof(int) * n);
      int * P_abaixo_y = (int *) malloc (sizeof(int) * n);
      int * P_acima_x = (int *) malloc (sizeof(int) * n);
      int * P_acima_y = (int *) malloc (sizeof(int) * n);
      int * S_abaixo;
      int * S_acima;
      
      int i, n_abaixo, n_acima, n_med, meio;

      v = (noArvoreIntervalosVIndireto *) malloc (sizeof(noArvoreIntervalosVIndireto));

      meio = (int) (n / 2.0 + 0.5);
      
      S_abaixo = (int *) malloc (sizeof(int) * meio); 
      S_acima = (int *) malloc (sizeof(int) * meio);
      
      v->p_med = S[meio-1]; /*Isso eh um segmento mas tenho que considerar o seu ponto extremo esquerdo.*/

      i = 0, n_abaixo = 0, n_med = 0;

      while (i < n && vetorGlobal[S[i]].get_pe().getY() <= vetorGlobal[v->p_med].get_pe().getY()){
	    if (vetorGlobal[S[i]].get_pd().getY() < vetorGlobal[v->p_med].get_pe().getY()){
	      S_abaixo[n_abaixo] = S[i];
	      n_abaixo++;
	    }
	    else{
	      P_acima_x[n_med] = S[i];
	      P_acima_y[n_med] = S[i];
	      P_abaixo_x[n_med] = S[i];/*P_ex já está ordenado.*/
	      P_abaixo_y[n_med] = S[i];
	      n_med++;
	    }
	    i++;
      }

      n_acima = 0;

      while (i < n){
	    S_acima[n_acima] = S[i];
	    n_acima++;
	    i++; 
      }
      
      /*Lembre que P_abaixo_y ja esta ordenado. Isso vale porque S está ordenado pela y-coordenada dos pontos extremos esq.*/
      sort(P_abaixo_x, P_abaixo_x + n_med, compare_pe_x);
      /*Lembre que S_med_pe tmb ja esta ordenado pelo msm motivo.*/
           
      v->arvoreymS = ArvorePrioridadeymSIndireto::ConstroiArvorePrioridadeymS(vetorGlobal, P_abaixo_x, P_abaixo_y, n_med, ponto_e);

      sort(P_acima_x, P_acima_x + n_med, compare_pd_x);
      sort(P_acima_y, P_acima_y + n_med, compare_pd_y);
      
      v->arvoreyMS = ArvorePrioridadeyMSIndireto::ConstroiArvorePrioridadeyMS(vetorGlobal, P_acima_x, P_acima_y, n_med, ponto_d);

      v->abaixo = ConstroiArvoreIntervalosVIndireto(vetorGlobal, S_abaixo, n_abaixo);
      v->acima = ConstroiArvoreIntervalosVIndireto(vetorGlobal, S_acima, n_acima);      
      free(P_abaixo_x);
      free(P_abaixo_y);
      free(P_acima_x);
      free(P_acima_y);
      free(S_abaixo);
      free(S_acima);
    }
    return v;
  }
  
  static void ConsultaArvoreIntervalosVIndireto (Segmento * vetorGlobal, noArvoreIntervalosVIndireto * v, Segmento Seg, list <int> &l){
    
    Janela W_ilim;
    if (v != NULL){
      //if (Seg.get_pe().yMaior(v->p_med)){
      if (Seg.get_pe().getY() > vetorGlobal[v->p_med].get_pe().getY()){
	    W_ilim = Janela(Seg.get_pe(), Ponto(Seg.get_pd().getX(), MaxInteger));
	    ArvorePrioridadeyMSIndireto::ConsultaArvorePrioridadeyMS(vetorGlobal, *(v->arvoreyMS), W_ilim, l);
	    ConsultaArvoreIntervalosVIndireto (vetorGlobal, v->acima, Seg, l);
      }
      else{
	    W_ilim = Janela(Ponto(Seg.get_pe().getX(), MinInteger), Seg.get_pd());
	    ArvorePrioridadeymSIndireto::ConsultaArvorePrioridadeymS(vetorGlobal, *(v->arvoreymS), W_ilim, l);
	    ConsultaArvoreIntervalosVIndireto (vetorGlobal, v->abaixo, Seg, l);
      }
    }
  }
  
  static void ImprimeArvoreIntervalosV (Segmento * vetorGlobal, noArvoreIntervalosVIndireto * v){
    if (v != NULL){
      printf("\n");
      Ponto::ImprimePonto(vetorGlobal[v->p_med].get_pe());
      printf("\nW-:\n");
      ArvorePrioridadeymSIndireto::ImprimeArvorePrioridadeymS(vetorGlobal, *(v->arvoreymS));
      printf("\nW+:\n");
      ArvorePrioridadeyMSIndireto::ImprimeArvorePrioridadeyMS(vetorGlobal, *(v->arvoreyMS));
      ImprimeArvoreIntervalosV (vetorGlobal, v->abaixo);
      ImprimeArvoreIntervalosV (vetorGlobal, v->acima);
    }
  }

  static void inicializa (noArvoreIntervalosVIndireto * T){
    noArvoreIntervalosVIndireto * v = T;
    if (v != NULL){
      ArvorePrioridadeymSIndireto::inicializa(v->arvoreymS);
      ArvorePrioridadeyMSIndireto::inicializa(v->arvoreyMS);
      ArvoreIntervalosVIndireto::inicializa(v->abaixo);
      ArvoreIntervalosVIndireto::inicializa(v->acima);
      free(v);
    }
  }
  
  static size_t consumo_espaco (noArvoreIntervalosVIndireto * T){
    size_t c = 0;
    noArvoreIntervalosVIndireto * v = T;
    if (v != NULL){
      c = sizeof(v) + sizeof(v->p_med);
      c += ArvorePrioridadeymSIndireto::consumo_espaco(v->arvoreymS);
      c += ArvorePrioridadeyMSIndireto::consumo_espaco(v->arvoreyMS);
      c += ArvoreIntervalosVIndireto::consumo_espaco(v->abaixo);
      c += ArvoreIntervalosVIndireto::consumo_espaco(v->acima);
    }
    return c;
  }
};
