struct noArvoreLimiteAIndireto {
  int p;
  int pDireita;
  noArvoreLimiteAIndireto * esq;
  noArvoreLimiteAIndireto *dir;
};

struct noArvoreLimiteIndireto {
  int p;
  noArvoreLimiteIndireto * esq;
  noArvoreLimiteIndireto * dir;
  noArvoreLimiteAIndireto * T_assoc;
};

class ArvoreLimiteIndireto {
    
 private:
  static noArvoreLimiteAIndireto * ConstroiArvorePerfeitamenteBalanceada(int * vet, int tam){
    int h = 0;
    int numNoH = 1; 
    int numNoTotal = numNoH;

    while(numNoH < tam){
      h++;
      numNoH *= 2;
    }

    /*numeros de folhas no último e no penúltimo nível.*/
    int X, Y;
    
    Y = numNoH - tam;
    X = tam - Y;

    //numNoTotal = X + X + Y + Y - 1;
    numNoTotal = (2*tam)-1;

    noArvoreLimiteAIndireto * arvore;
    arvore = (noArvoreLimiteAIndireto *) malloc (sizeof(noArvoreLimiteAIndireto) * numNoTotal);
    
    if (arvore == NULL){
      printf("Erro na alocação de memória para uma árvore limite associada.\n");
      exit(0);
    }

    int i = numNoTotal-1;
    
    /*Construção das folhas do último nível*/
    for (int j = X-1; j >= 0; j--){
      arvore[i].p = vet[j];
      arvore[i].pDireita = vet[j];
      arvore[i].esq = NULL;
      arvore[i].dir = NULL;
      i--;
    }
    
    /*Construção das folhas do penúltimo nível*/
    for (int j = tam-1; j >= X; j--){
      arvore[i].p = vet[j];
      arvore[i].pDireita = vet[j];
      arvore[i].esq = NULL;
      arvore[i].dir = NULL;
      i--;
    }
    
    /*Construção dos nós internos*/
    /*Ainda tenho que provar que, pela construção, todo nó interno tem um filho à esquerda e à direita*/
    int e, d;//filhos à esquerda (com íncide igual à e) e à direita (com índice igual à d) de arvore[i].
    //for (int j = 0; j < numNosInternos; j++){
    while (i >= 0){
      e = i*2+1;
      d = i*2+2;
      arvore[i].p = arvore[e].pDireita;
      arvore[i].pDireita = arvore[d].pDireita;
      arvore[i].esq = &(arvore[e]);
      arvore[i].dir = &(arvore[d]);
      i--;
    }

    return arvore;
  }

 public:
  ArvoreLimiteIndireto(){
  }
  

  //vetOrdenadoX[0 .. tam-1] e vetOrdenadoY[0 .. tam-1].
  static noArvoreLimiteIndireto * ConstroiArvoreLimiteIndireto(Segmento * vetorGlobal, int * vetX, int * vetY, int tam, char * p){
    
    noArvoreLimiteIndireto * v;
    v = (noArvoreLimiteIndireto *) malloc (sizeof(noArvoreLimiteIndireto));
    v->T_assoc = ArvoreLimiteIndireto::ConstroiArvorePerfeitamenteBalanceada(vetY, tam);
    
    int i = (int)((tam/2.0)+0.5);
      
    /**************x-vetor - esquerda e direita************/
    int * novoVetOrdenadoXE = (int *) malloc (sizeof(int)*i);
    int k;
    for (k = 0; k < i; k++)
      novoVetOrdenadoXE[k] = vetX[k];
    
    int * novoVetOrdenadoXD = (int *) malloc (sizeof(int)*(tam-i));
    int indxd = 0;
    for (k = i; k < tam; k++){
      novoVetOrdenadoXD[indxd] = vetX[k];
      indxd++;
    }

    /**************y-vetor - esquerda e direita************/
    int * novoVetOrdenadoYE = (int *) malloc (sizeof(int)*i);
    int * novoVetOrdenadoYD = (int *) malloc (sizeof(int)*(tam-i));

    int indexe = 0;
    int indexd = 0;
    for (k = 0; k < tam; k++){
      
      if (vetorGlobal[vetY[k]].qualPonto(p).xMenorIgual(vetorGlobal[vetX[i-1]].qualPonto(p))){
	    novoVetOrdenadoYE[indexe] = vetY[k];
	    indexe++;
      }
      else{
	    novoVetOrdenadoYD[indexd] = vetY[k];
	    indexd++;
      }
    }
    
    /*******************************************************/

    v->p = vetX[i-1];
    
    if (tam <= 1){
      v->esq = NULL;
      v->dir = NULL;
    }
    else{
      v->esq = ConstroiArvoreLimiteIndireto(vetorGlobal, novoVetOrdenadoXE, novoVetOrdenadoYE, i, p);
      v->dir = ConstroiArvoreLimiteIndireto(vetorGlobal, novoVetOrdenadoXD, novoVetOrdenadoYD, tam - i, p);
    }
    free(novoVetOrdenadoXE);
    free(novoVetOrdenadoYE);
    free(novoVetOrdenadoXD);
    free(novoVetOrdenadoYD);
    return v;
  }
  

  //Recebe uma árvore limite, o ponto inferior esquerdo e superior direito de uma janela W.
  //Devolve um nó da árvore limite dentro de uma janela.
  static noArvoreLimiteAIndireto * EncontraNoDivideY (Segmento * vetorGlobal, noArvoreLimiteAIndireto *v, Ponto p_ie, Ponto p_sd, char * p){
    noArvoreLimiteAIndireto  * ret = v;
    //enquanto ret não é uma folha e ret.p não está em um intervalo de W.
    while (!(ArvoreLimiteIndireto::isFolha(ret)) && (p_sd.yMenorIgual(vetorGlobal[ret->p].qualPonto(p)) || p_ie.yMaior(vetorGlobal[ret->p].qualPonto(p)))){
      if (p_sd.yMenorIgual(vetorGlobal[ret->p].qualPonto(p)))
	    ret = (ret->esq);
      else
	    ret = (ret->dir);
    }
    return ret;
  }
  
  //Recebe uma árvore limite, o ponto inferior esquerdo e superior direito de uma janela W.
  //Devolve um nó da árvore limite dentro de uma janela.
  static noArvoreLimiteIndireto * EncontraNoDivideX (Segmento * vetorGlobal, noArvoreLimiteIndireto *v, Ponto p_ie, Ponto p_sd, char * p){
    noArvoreLimiteIndireto  * ret = v;
    //enquanto ret não é uma folha e ret.p não está em um intervalo de W.
    while (!(ArvoreLimiteIndireto::isFolha(ret)) && (p_sd.xMenorIgual(vetorGlobal[ret->p].qualPonto(p)) || p_ie.xMaior(vetorGlobal[ret->p].qualPonto(p)))){
      if (p_sd.xMenorIgual(vetorGlobal[ret->p].qualPonto(p)))
 	    ret = (ret->esq);
      else
	    ret = (ret->dir);
    }
    return ret;
  }

  static void ConsultaArvoreLimiteIndireto1D(Segmento * vetorGlobal, noArvoreLimiteAIndireto * v, Janela W, list <int> & l, char * p){
    //list <Ponto> l;
   
    noArvoreLimiteAIndireto * v_div = ArvoreLimiteIndireto::EncontraNoDivideY(vetorGlobal, v, W.getPontoInferiorEsquerdo(), W.getPontoSuperiorDireito(), p);
    
    if (ArvoreLimiteIndireto::isFolha(v_div)){//se v_div é uma folha.
      if (W.contem(vetorGlobal[v_div->p].qualPonto(p))){
	    l.push_back(v_div->p);
      }
    }
    else{
      //andando para a esquerda da árvore limite com raiz em *v.
      noArvoreLimiteAIndireto * v_aux = v_div->esq;
      while(!(ArvoreLimiteIndireto::isFolha(v_aux))){//enquanto v_aux não é uma folha.
	    if (W.getPontoInferiorEsquerdo().yMenorIgual(vetorGlobal[v_aux->p].qualPonto(p))){
	      ArvoreLimiteIndireto::PontosSubarvore(v_aux->dir, l);
	      v_aux = v_aux->esq;
	    }
	    else{
	      v_aux = v_aux->dir;
 	    }
      } 

      if (W.contem(vetorGlobal[v_aux->p].qualPonto(p))){//verifica se o ponto armazenado em uma folha está em W.
	    l.push_back(v_aux->p);
      }
      //andando para a direita da árvore limite com raiz em *v.
      v_aux = v_div->dir;
      while(!(ArvoreLimiteIndireto::isFolha(v_aux))){//enquanto v_aux não é uma folha.
	    if (W.getPontoSuperiorDireito().yMaior(vetorGlobal[v_aux->p].qualPonto(p))){
	      ArvoreLimiteIndireto::PontosSubarvore(v_aux->esq, l);
	      v_aux = v_aux->dir;
	    }
	    else{
	      v_aux = v_aux->esq;
	    }
      }
      if (W.contem(vetorGlobal[v_aux->p].qualPonto(p))){//verifica se o ponto armazenado em uma folha está em W.
	    l.push_back(v_aux->p);
      }
    }
  }
  
  static void ConsultaArvoreLimiteIndireto(Segmento * vetorGlobal, noArvoreLimiteIndireto * v, Janela W, list <int> & l, char * p){
    /*list <int> l;*/
    noArvoreLimiteIndireto * v_div = ArvoreLimiteIndireto::EncontraNoDivideX(vetorGlobal, v, W.getPontoInferiorEsquerdo(), W.getPontoSuperiorDireito(), p);
    if (ArvoreLimiteIndireto::isFolha(v_div)){//se v_div é uma folha.
      if (W.contem(vetorGlobal[v_div->p].qualPonto(p))){
	    l.push_back(v_div->p);
      }
    }
    else{
      //andando para a esquerda da árvore limite com raiz em *v.
      noArvoreLimiteIndireto * v_aux = v_div->esq;
      
      while(!(ArvoreLimiteIndireto::isFolha(v_aux))){//enquanto v_aux não é uma folha.

	    if (W.getPontoInferiorEsquerdo().xMenorIgual(vetorGlobal[v_aux->p].qualPonto(p))){
	      ArvoreLimiteIndireto::ConsultaArvoreLimiteIndireto1D(vetorGlobal, v_aux->dir->T_assoc, W, l, p);
	      v_aux = v_aux->esq;
	    }
	    else{
	      v_aux = v_aux->dir;
	    }
      }
     
      if (W.contem(vetorGlobal[v_aux->p].qualPonto(p))){//verifica se o ponto armazenado em uma folha está em W.
	    l.push_back(v_aux->p);
      }
      //andando para a direita da árvore limite com raiz em *v.
      v_aux = v_div->dir;
      while(!(ArvoreLimiteIndireto::isFolha(v_aux))){//enquanto v_aux não é uma folha.
	    if (W.getPontoSuperiorDireito().xMaior(vetorGlobal[v_aux->p].qualPonto(p))){
	      ArvoreLimiteIndireto::ConsultaArvoreLimiteIndireto1D(vetorGlobal, v_aux->esq->T_assoc, W, l, p);
	      v_aux = v_aux->dir;
	    }
	    else{
	      v_aux = v_aux->esq;
	    }
      }
      if (W.contem(vetorGlobal[v_aux->p].qualPonto(p))){//verifica se o ponto armazenado em uma folha está em W.
	    l.push_back(v_aux->p);
      }
    }
  }


static void PontosSubarvore (noArvoreLimiteAIndireto * v, list <int> &l) {
   
    if (ArvoreLimiteIndireto::isFolha(v)){
      l.push_back(v->p);
    }
    else{
      PontosSubarvore(v->esq, l);
      PontosSubarvore(v->dir, l);
    }
  }


static void inicializa (noArvoreLimiteIndireto * T){
	noArvoreLimiteIndireto * v = T;
	if (v != NULL){
	  ArvoreLimiteIndireto::inicializa(v->esq);
	  ArvoreLimiteIndireto::inicializa(v->dir);
	  ArvoreLimiteIndireto::inicializaAssociada(v->T_assoc);
	  free(v);  
	}
}

static void inicializaAssociada (noArvoreLimiteAIndireto * T){
  noArvoreLimiteAIndireto * v = T;
  if (v != NULL){
  	free(v);
  }
}

static size_t consumo_espaco (noArvoreLimiteIndireto *T){
  size_t c = 0;
  noArvoreLimiteIndireto * v = T;
  if (v != NULL){
    c = sizeof (v) + sizeof(v->p);
    c += ArvoreLimiteIndireto::consumo_espaco_Associada(v->T_assoc);
    c += ArvoreLimiteIndireto::consumo_espaco (v->esq);
    c += ArvoreLimiteIndireto::consumo_espaco (v->dir);
  }
  return c;
}

static size_t consumo_espaco_Associada (noArvoreLimiteAIndireto *T){
  size_t c = 0;
  noArvoreLimiteAIndireto * v = T;
  if (v != NULL){
    c = sizeof(v) + sizeof(v->p) + sizeof(v->pDireita);
    c += ArvoreLimiteIndireto::consumo_espaco_Associada (v->esq);
    c += ArvoreLimiteIndireto::consumo_espaco_Associada (v->dir);
  }
  return c;
}

static void ImprimeArvore(Segmento * vetorGlobal, noArvoreLimiteIndireto * T){
  noArvoreLimiteIndireto * v = T;
  if (v != NULL){
    Ponto::ImprimePonto(vetorGlobal[v->p].get_pe());
    printf("\n");
    ArvoreLimiteIndireto::ImprimeArvoreAssociada(vetorGlobal, v->T_assoc);
    ArvoreLimiteIndireto::ImprimeArvore(vetorGlobal, v->esq);
    ArvoreLimiteIndireto::ImprimeArvore(vetorGlobal, v->dir);
  } 
}

static void ImprimeArvoreAssociada(Segmento * vetorGlobal, noArvoreLimiteAIndireto * T){
	noArvoreLimiteAIndireto * v = T;
	if (v != NULL){
      Ponto::ImprimePonto(vetorGlobal[v->p].get_pe());
      printf(" ");
      Ponto::ImprimePonto(vetorGlobal[v->pDireita].get_pd());
      ArvoreLimiteIndireto::ImprimeArvoreAssociada(vetorGlobal, v->esq);
      ArvoreLimiteIndireto::ImprimeArvoreAssociada(vetorGlobal, v->dir);
	}
}

  template<class myType>
  static bool isFolha(myType * v){
    return (v->esq == NULL);
  }
};

