/*************************************************************************************************************************
  Seja P um conjunto de pontos. Seja W uma janela com w_1 = (x, y) e w_2 = (+00, y').
  Queremos encontrar os pontos de P que pertencem a W.
  Obs.: Uma arvore de busca com prioridade desta classe tem armazenado também segmentos.
 ************************************************************************************************************************/
struct noArvorePrioridadexMSIndireto {
  int p_max;
  int p_med;
  struct noArvorePrioridadexMSIndireto * acima;
  struct noArvorePrioridadexMSIndireto * abaixo;
};

class ArvorePrioridadexMSIndireto {
 public:
  /*
    Recebe um conjunto P de pontos representados por dois vetores P_x (x-ordenado) e P_y (y-ordenado) e um vetor de 
    segmentos S (ordenado de forma crescente (x-coordenada) por seus pontos extremos direitos) cujo pontos extremos
    direitos são os pontos de P_x e P_y.
    Devolve uma arvore de busca com prioridade.
  */
  static noArvorePrioridadexMSIndireto * ConstroiArvorePrioridadexMS (Segmento * vetorGlobal, int * Px, int * Py, int tamanho, char * ponto){
    noArvorePrioridadexMSIndireto * v;
    
    if (tamanho > 0){
      v = (noArvorePrioridadexMSIndireto *) malloc (sizeof(noArvorePrioridadexMSIndireto));
      int meio = (int) (((tamanho-1)/2.0)+0.5);
      int deslocamento = 0;
      int i, indexy, indexx_ac, indexx_ab;
      
      int * Py_abaixo = (int *) malloc (sizeof(int) * meio);
      int * Py_acima = (int *) malloc (sizeof(int) * (tamanho - 1 - meio));
      int * Px_abaixo = (int *) malloc (sizeof(int) * meio);
      int * Px_acima = (int *) malloc (sizeof(int) * (tamanho - 1 - meio));
      
      v->p_max = Px[tamanho-1];
      
      for (i = 0, indexy = 0; i < meio + deslocamento; i++)
	    if (!(vetorGlobal[Py[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_max].qualPonto(ponto)))){
 	      Py_abaixo[indexy] = Py[i];
 	      indexy++;
	    }
	    else
	      deslocamento = 1;
      indexy = 0;
      while (i < tamanho){
	    if (!(vetorGlobal[Py[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_max].qualPonto(ponto)))){
	      Py_acima[indexy] = Py[i];
	      indexy++;
	    }
	    i++;
      }
      
      v->p_med = Py[meio-1+deslocamento];
      
      for (i = 0, indexx_ab = 0, indexx_ac = 0; i < tamanho-1; i++){
	    if (vetorGlobal[Px[i]].qualPonto(ponto).yMenorIgual(vetorGlobal[v->p_med].qualPonto(ponto))){
	      Px_abaixo[indexx_ab] = Px[i];
	      indexx_ab++;
	    }
	    else{
	      Px_acima[indexx_ac] = Px[i];
	      indexx_ac++;
	    }
      }

      v->abaixo = ConstroiArvorePrioridadexMS (vetorGlobal, Px_abaixo, Py_abaixo, meio, ponto);
      v->acima = ConstroiArvorePrioridadexMS (vetorGlobal, Px_acima, Py_acima, tamanho - 1 - meio, ponto);
      free(Py_abaixo);
      free(Py_acima);
      free(Px_abaixo);
      free(Px_acima);

    }
    else
      v = NULL;
    return v;
  }
  
  static void ConsultaArvorePrioridadexMS (Segmento * vetorGlobal, noArvorePrioridadexMSIndireto v, Janela W, list <int> &l){
        
    noArvorePrioridadexMSIndireto u = v;
    noArvorePrioridadexMSIndireto uLinha;
    /*Pontos w1 e w2 de uma janela ilimitada no eixo x para -\infty.*/
    Ponto w1 = W.getPontoInferiorEsquerdo();
    Ponto w2 = W.getPontoSuperiorDireito();
    
    while (!(ArvorePrioridadexMSIndireto::isFolha(u)) && !(ArvorePrioridadexMSIndireto::isQuaseFolha(u)) 
	   && (w1.yMaior(vetorGlobal[u.p_med].get_pd()) || w2.yMenorIgual(vetorGlobal[u.p_med].get_pd()))){
            
      /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
      if (W.contem(vetorGlobal[u.p_max].get_pd())){
	    l.push_back(u.p_max);
      }
      
      if (w1.yMaior(vetorGlobal[u.p_med].get_pd()))
	    u = *(u.acima);
      else
	    u = *(u.abaixo);
    }
   

    if(!(ArvorePrioridadexMSIndireto::isFolha(u)) && !(ArvorePrioridadexMSIndireto::isQuaseFolha(u))){

      /*Verifica se u=v_div está em W.*/
      if (W.contem(vetorGlobal[u.p_max].get_pd())){
	    l.push_back(u.p_max);
      }

      /*Ande para baixo de u.*/
      uLinha = *(u.abaixo);
      while (!(ArvorePrioridadexMSIndireto::isFolha(uLinha)) && !(ArvorePrioridadexMSIndireto::isQuaseFolha(uLinha))){
      
        /*Neste ponto verificamos se o ponto armazenado no no' uLinha esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	      l.push_back(uLinha.p_max);
	    }
	
	    if (w1.yMenorIgual(vetorGlobal[uLinha.p_med].get_pd())){
	      ArvorePrioridadexMSIndireto::PontosSubarvore(vetorGlobal, uLinha.acima, W, l);
	      uLinha = *(uLinha.abaixo);
	    }
    	else
	      uLinha = *(uLinha.acima);
      }
      
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	    l.push_back(uLinha.p_max);
      }
      if (ArvorePrioridadexMSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_max(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pd())){
	      l.push_back(uLinha.abaixo->p_max);
	    }
      }
      
      /*Ande para cima de u.*/
      uLinha = *(u.acima);
      while (!(ArvorePrioridadexMSIndireto::isFolha(uLinha)) && !(ArvorePrioridadexMSIndireto::isQuaseFolha(uLinha))){
	
	    /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	      l.push_back(uLinha.p_max);
	    }

	    if (w2.yMaior(vetorGlobal[uLinha.p_med].get_pd())){
	      ArvorePrioridadexMSIndireto::PontosSubarvore(vetorGlobal, uLinha.abaixo, W, l);
	      uLinha = *(uLinha.acima);
	    }
	    else
	      uLinha = *(uLinha.abaixo);
      }
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	    l.push_back(uLinha.p_max);
      }
      if (ArvorePrioridadexMSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_max(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pd())){
	      l.push_back(uLinha.abaixo->p_max);
	    }
      }
    }
    /*Neste ponto u eh ou uma folha ou uma quase folha.*/
    else {
      if (W.contem(vetorGlobal[u.p_max].get_pd())){
	    l.push_back(u.p_max);
      }
      if (ArvorePrioridadexMSIndireto::isQuaseFolha(u)){
	    /*Neste ponto u eh quase folha. Então p_max(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[u.p_med].get_pd())){
	      l.push_back(u.abaixo->p_max);
	    }
      }
    }
  }
  
  static void PontosSubarvore (Segmento * vetorGlobal, noArvorePrioridadexMSIndireto * v, Janela W, list <int> & l){
    if (v != NULL){
      if(vetorGlobal[v->p_max].get_pd().xMaior(W.getPontoInferiorEsquerdo()) || vetorGlobal[v->p_max].get_pd().Igual(W.getPontoInferiorEsquerdo())){
        l.push_back(v->p_max);
	    PontosSubarvore (vetorGlobal, v->abaixo, W, l);
	    PontosSubarvore (vetorGlobal, v->acima, W, l);
      }
    }
  }

  static bool isFolha(noArvorePrioridadexMSIndireto v){
    return (v.acima == NULL && v.abaixo == NULL);
  }

  static bool isQuaseFolha(noArvorePrioridadexMSIndireto v){
    return (v.abaixo != NULL && v.acima == NULL);
  }

   static void ImprimeArvorePrioridadexMS (Segmento * vetorGlobal, noArvorePrioridadexMSIndireto v){
    Ponto::ImprimePonto(vetorGlobal[v.p_max].get_pd());printf(" -> ");
    Segmento::ImprimeSegmento(vetorGlobal[v.p_max]);printf(" ");
    Ponto::ImprimePonto(vetorGlobal[v.p_med].get_pd());
    if (v.abaixo != NULL)
      ImprimeArvorePrioridadexMS(vetorGlobal, *v.abaixo);
    if (v.acima != NULL)
      ImprimeArvorePrioridadexMS(vetorGlobal, *v.acima);
  }

 static void inicializa (noArvorePrioridadexMSIndireto * T){
    noArvorePrioridadexMSIndireto * v = T;
    if (v != NULL){
      ArvorePrioridadexMSIndireto::inicializa(v->abaixo);
      ArvorePrioridadexMSIndireto::inicializa(v->acima);
      free(v);
    }
  }
  
  static size_t consumo_espaco (noArvorePrioridadexMSIndireto * T){
  	size_t c = 0;
    noArvorePrioridadexMSIndireto * v = T; 
    if (v != NULL){
      c = sizeof (v) + sizeof(v->p_med) + sizeof(v->p_max);
      c += ArvorePrioridadexMSIndireto::consumo_espaco(v->abaixo);
      c += ArvorePrioridadexMSIndireto::consumo_espaco(v->acima);
    }
    return c;
  }
  
};
