/*************************************************************************************************************************
  Seja P um conjunto de pontos. Seja W uma janela com w_1 = (-00, y) e w_2 = (x', y').
  Queremos encontrar os pontos de P que pertencem a W.
  Obs.: Uma arvore de busca com prioridade desta classe tem armazenado também segmentos.
 ************************************************************************************************************************/
struct noArvorePrioridadexmSIndireto {
  int p_min;
  int p_med;
  struct noArvorePrioridadexmSIndireto * acima;
  struct noArvorePrioridadexmSIndireto * abaixo;
};

class ArvorePrioridadexmSIndireto {
 public:
  /*****************************************************************************************************************
    Recebe um conjunto P de pontos representados por dois vetores P_x (x-ordenado) e P_y (y-ordenado) e um vetor de 
    segmentos S (ordenado de forma crescente (x-coordenada) por seus pontos extremos direitos) cujo pontos extremos
    esquerdos são os pontos de P_x e P_y.
    Devolve uma arvore de busca com prioridade.
  ******************************************************************************************************************/
  static noArvorePrioridadexmSIndireto * ConstroiArvorePrioridadexmS (Segmento * vetorGlobal, int * Px, int * Py, int tamanho, char * ponto){
    noArvorePrioridadexmSIndireto * v;
    
    if (tamanho > 0){
      v = (noArvorePrioridadexmSIndireto *) malloc (sizeof(noArvorePrioridadexmSIndireto));
      int meio = (int) (((tamanho-1)/2.0)+0.5);
      int deslocamento = 0;
      int i, indexy, indexx_ab, indexx_ac;
      
      int * Py_abaixo = (int *) malloc (sizeof(int) * meio);
      int * Py_acima = (int *) malloc (sizeof(int) * (tamanho - 1 - meio));
      int * Px_abaixo = (int *) malloc (sizeof(int) * meio);
      int * Px_acima = (int *) malloc (sizeof(int) * (tamanho - 1 - meio));
      
      v->p_min = Px[0];
            
      for (i = 0, indexy = 0; i < meio + deslocamento; i++)
	    if (!(vetorGlobal[Py[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_min].qualPonto(ponto)))){
	      Py_abaixo[indexy] = Py[i];
	      indexy++;
	    }
	    else
	      deslocamento = 1;
      indexy = 0;
      while (i < tamanho){
	    if (!(vetorGlobal[Py[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_min].qualPonto(ponto)))){
	      Py_acima[indexy] = Py[i];
	      indexy++;
	    }
	    i++;
      }
      
      v->p_med = Py[meio-1+deslocamento];
      
      for (i = 1, indexx_ab = 0, indexx_ac = 0; i < tamanho; i++){
	    if (vetorGlobal[Px[i]].qualPonto(ponto).yMenorIgual(vetorGlobal[v->p_med].qualPonto(ponto))){
	      Px_abaixo[indexx_ab] = Px[i];
	      indexx_ab++;
	    }
	    else{
	      Px_acima[indexx_ac] = Px[i];
	      indexx_ac++;
	    }
      }

      v->abaixo = ConstroiArvorePrioridadexmS (vetorGlobal, Px_abaixo, Py_abaixo, meio, ponto);
      v->acima = ConstroiArvorePrioridadexmS (vetorGlobal, Px_acima, Py_acima, tamanho - 1 - meio, ponto);
      free(Py_abaixo);
      free(Py_acima);
      free(Px_abaixo);
      free(Px_acima);
    }
    else
      v = NULL;
    return v;
  }

static void ConsultaArvorePrioridadexmS (Segmento * vetorGlobal, noArvorePrioridadexmSIndireto v, Janela W, list <int> &l){
        
    noArvorePrioridadexmSIndireto u = v;
    noArvorePrioridadexmSIndireto uLinha;
    /*Pontos w1 e w2 de uma janela ilimitada no eixo x para -\infty.*/
    Ponto w1 = W.getPontoInferiorEsquerdo();
    Ponto w2 = W.getPontoSuperiorDireito();

    while (!(ArvorePrioridadexmSIndireto::isFolha(u)) && !(ArvorePrioridadexmSIndireto::isQuaseFolha(u)) 
	   && (w1.yMaior(vetorGlobal[u.p_med].get_pe()) || w2.yMenorIgual(vetorGlobal[u.p_med].get_pe()))){
            
      /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
      if (W.contem(vetorGlobal[u.p_min].get_pe())){
	    l.push_back(u.p_min);
      }
      
      if (w1.yMaior(vetorGlobal[u.p_med].get_pe()))
	    u = *(u.acima);
      else
	    u = *(u.abaixo);
    }
   

    if(!(ArvorePrioridadexmSIndireto::isFolha(u)) && !(ArvorePrioridadexmSIndireto::isQuaseFolha(u))){

      /*Verifica se u=v_div está em W.*/
      if (W.contem(vetorGlobal[u.p_min].get_pe())){
	    l.push_back(u.p_min);
      }

      /*Ande para baixo de u.*/
      uLinha = *(u.abaixo);
      while (!(ArvorePrioridadexmSIndireto::isFolha(uLinha)) && !(ArvorePrioridadexmSIndireto::isQuaseFolha(uLinha))){
      
        /*Neste ponto verificamos se o ponto armazenado no no' uLinha esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	      l.push_back(uLinha.p_min);
	    }
	 
	    if (w1.yMenorIgual(vetorGlobal[uLinha.p_med].get_pe())){
	      ArvorePrioridadexmSIndireto::PontosSubarvore(vetorGlobal, uLinha.acima, W, l);
	      uLinha = *(uLinha.abaixo);
	    }
	    else
	      uLinha = *(uLinha.acima);
      } 
      
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	    l.push_back(uLinha.p_min);
      }
      if (ArvorePrioridadexmSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_min(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pe())){
	      l.push_back(uLinha.abaixo->p_min);
	    }
      }
      
      /*Ande para cima de u.*/
      uLinha = *(u.acima);
      while (!(ArvorePrioridadexmSIndireto::isFolha(uLinha)) && !(ArvorePrioridadexmSIndireto::isQuaseFolha(uLinha))){
	
	    /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	      l.push_back(uLinha.p_min);
	    }

	    if (w2.yMaior(vetorGlobal[uLinha.p_med].get_pe())){
	      ArvorePrioridadexmSIndireto::PontosSubarvore(vetorGlobal, uLinha.abaixo, W, l);
	      uLinha = *(uLinha.acima);
	    }
	    else
	      uLinha = *(uLinha.abaixo);
      }
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	    l.push_back(uLinha.p_min);
      }
      if (ArvorePrioridadexmSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_min(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pe())){
	      l.push_back(uLinha.abaixo->p_min);
	    }
      }
    }
    /*Neste ponto u eh ou uma folha ou uma quase folha.*/
    else {
      if (W.contem(vetorGlobal[u.p_min].get_pe())){
	    l.push_back(u.p_min);
      }
      if (ArvorePrioridadexmSIndireto::isQuaseFolha(u)){
	    /*Neste ponto u eh quase folha. Então p_min(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[u.p_med].get_pe())){
	      l.push_back(u.abaixo->p_min);
	    }
      }
    }
  }  
  
  static void PontosSubarvore (Segmento * vetorGlobal, noArvorePrioridadexmSIndireto * v, Janela W, list <int> & l){
    if (v != NULL){
      if(vetorGlobal[v->p_min].get_pe().xMenorIgual(W.getPontoSuperiorDireito())){
        l.push_back(v->p_min);
	    PontosSubarvore (vetorGlobal, v->abaixo, W, l);
	    PontosSubarvore (vetorGlobal, v->acima, W, l);
      }
    }
  }

  static bool isFolha(noArvorePrioridadexmSIndireto v){
    return (v.acima == NULL && v.abaixo == NULL);
  }

  static bool isQuaseFolha(noArvorePrioridadexmSIndireto v){
    return (v.abaixo != NULL && v.acima == NULL);
  }

   static void ImprimeArvorePrioridadexmS (Segmento * vetorGlobal, noArvorePrioridadexmSIndireto v){
    Ponto::ImprimePonto(vetorGlobal[v.p_min].get_pe());printf(" -> ");
    Segmento::ImprimeSegmento(vetorGlobal[v.p_min]);printf(" ");
    Ponto::ImprimePonto(vetorGlobal[v.p_med].get_pe());
    if (v.abaixo != NULL)
      ImprimeArvorePrioridadexmS(vetorGlobal, *v.abaixo);
    if (v.acima != NULL)
      ImprimeArvorePrioridadexmS(vetorGlobal, *v.acima);
  }

  static void inicializa (noArvorePrioridadexmSIndireto * T){
    noArvorePrioridadexmSIndireto * v = T;
    if (v != NULL){
      ArvorePrioridadexmSIndireto::inicializa(v->abaixo);
      ArvorePrioridadexmSIndireto::inicializa(v->acima);
      free(v);
    }
  }
  
  static size_t consumo_espaco (noArvorePrioridadexmSIndireto * T){
  	size_t c = 0;
    noArvorePrioridadexmSIndireto * v = T; 
    if (v != NULL){
      c = sizeof(v) + sizeof(v->p_med) + sizeof(v->p_min);
      c += ArvorePrioridadexmSIndireto::consumo_espaco(v->abaixo);
      c += ArvorePrioridadexmSIndireto::consumo_espaco(v->acima);
    }
    return c;
  }
};
