/*************************************************************************************************************************
  Seja P um conjunto de pontos. Seja W uma janela com w_1 = (x, y) e w_2 = (x', +00).
  Queremos encontrar os pontos de P que pertencem a W.
  Obs.: Uma arvore de busca com prioridade desta classe tem armazenado também segmentos.
 ************************************************************************************************************************/
struct noArvorePrioridadeyMSIndireto {
  int p_max;
  int p_med;
  struct noArvorePrioridadeyMSIndireto * dir;
  struct noArvorePrioridadeyMSIndireto * esq;
};

class ArvorePrioridadeyMSIndireto {
 public:
  /*****************************************************************************************************************
    Recebe um conjunto P de pontos representados por dois vetores P_x (x-ordenado) e P_y (y-ordenado) e um vetor de 
    segmentos S (ordenado de forma crescente (y-coordenada) por seus pontos extremos direitos) cujo pontos extremos
    direitos são os pontos de P_x e P_y.
    Devolve uma arvore de busca com prioridade.
  ******************************************************************************************************************/
  static noArvorePrioridadeyMSIndireto * ConstroiArvorePrioridadeyMS (Segmento * vetorGlobal, int * Px, int * Py, int tamanho, char * ponto){
    noArvorePrioridadeyMSIndireto * v;
    
    if (tamanho > 0){
      v = (noArvorePrioridadeyMSIndireto *) malloc (sizeof(noArvorePrioridadeyMSIndireto));
      int meio = (int) (((tamanho-1)/2.0)+0.5);
      int deslocamento = 0;
      int i, indexx, indexy_e, indexy_d;
      
      int * Py_esq = (int *) malloc (sizeof(int)* meio);
      int * Py_dir = (int *) malloc (sizeof(int)* (tamanho - 1 - meio));
      int * Px_esq = (int *) malloc (sizeof(int)* meio);
      int * Px_dir = (int *) malloc (sizeof(int)* (tamanho - 1 - meio));
      
      v->p_max = Py[tamanho-1];
      
      for (i = 0, indexx = 0; i < meio + deslocamento; i++)
	    if (!(vetorGlobal[Px[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_max].qualPonto(ponto)))){
	      Px_esq[indexx] = Px[i];
	      indexx++;
	    } 
	    else
	      deslocamento = 1;
      indexx = 0;
      while (i < tamanho){
	    if (!(vetorGlobal[Px[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_max].qualPonto(ponto)))){
	      Px_dir[indexx] = Px[i];
	      indexx++;
	    }
	    i++;
      }
      
      v->p_med = Px[meio-1+deslocamento];
      
      for (i = 0, indexy_e = 0, indexy_d = 0; i < tamanho-1; i++){
	    if (vetorGlobal[Py[i]].qualPonto(ponto).xMenorIgual(vetorGlobal[v->p_med].qualPonto(ponto))){
	      Py_esq[indexy_e] = Py[i];
	      indexy_e++;
	    }
	    else{
   	      Py_dir[indexy_d] = Py[i];
   	      indexy_d++;
	    }
      }

      v->esq = ConstroiArvorePrioridadeyMS (vetorGlobal, Px_esq, Py_esq, meio, ponto);
      v->dir = ConstroiArvorePrioridadeyMS (vetorGlobal, Px_dir, Py_dir, tamanho - 1 - meio, ponto);
      free(Py_esq);
      free(Py_dir);
      free(Px_esq);
      free(Px_dir);
    }
    else
      v = NULL;
    return v;
  }
  
  static void ConsultaArvorePrioridadeyMS (Segmento * vetorGlobal, noArvorePrioridadeyMSIndireto v, Janela W, list <int> &l){
        
    noArvorePrioridadeyMSIndireto u = v;
    noArvorePrioridadeyMSIndireto uLinha;
    /*Pontos w1 e w2 de uma janela ilimitada no eixo x para -\infty.*/
    Ponto w1 = W.getPontoInferiorEsquerdo();
    Ponto w2 = W.getPontoSuperiorDireito();

    while (!(ArvorePrioridadeyMSIndireto::isFolha(u)) && !(ArvorePrioridadeyMSIndireto::isQuaseFolha(u)) 
	   && (w1.xMaior(vetorGlobal[u.p_med].get_pd()) || w2.xMenorIgual(vetorGlobal[u.p_med].get_pd()))){
            
      /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
      if (W.contem(vetorGlobal[u.p_max].get_pd())){
	    l.push_back(u.p_max);
      }
      
      if (w1.xMaior(vetorGlobal[u.p_med].get_pd()))
	    u = *(u.dir);
      else
	    u = *(u.esq);
    }
   

    if(!(ArvorePrioridadeyMSIndireto::isFolha(u)) && !(ArvorePrioridadeyMSIndireto::isQuaseFolha(u))){

      /*Verifica se u=v_div está em W.*/
      if (W.contem(vetorGlobal[u.p_max].get_pd())){
	    l.push_back(u.p_max);
      }

      /*Ande para a esquerda de u.*/
      uLinha = *(u.esq);
      while (!(ArvorePrioridadeyMSIndireto::isFolha(uLinha)) && !(ArvorePrioridadeyMSIndireto::isQuaseFolha(uLinha))){
      
        /*Neste ponto verificamos se o ponto armazenado no no' uLinha esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	      l.push_back(uLinha.p_max);
	    }
	
	    if (w1.xMenorIgual(vetorGlobal[uLinha.p_med].get_pd())){
	      ArvorePrioridadeyMSIndireto::PontosSubarvore(vetorGlobal, uLinha.dir, W, l);
	      uLinha = *(uLinha.esq);
	    } 
	    else
	      uLinha = *(uLinha.dir);
      }
      
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	    l.push_back(uLinha.p_max);
      }
      if (ArvorePrioridadeyMSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_max(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pd())){
	      l.push_back(uLinha.esq->p_max);
	    }
      }
      
      /*Ande para a direita de u.*/
      uLinha = *(u.dir);
      while (!(ArvorePrioridadeyMSIndireto::isFolha(uLinha)) && !(ArvorePrioridadeyMSIndireto::isQuaseFolha(uLinha))){
	
	    /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	      l.push_back(uLinha.p_max);
	    }

   	    if (w2.xMaior(vetorGlobal[uLinha.p_med].get_pd())){
	      ArvorePrioridadeyMSIndireto::PontosSubarvore(vetorGlobal, uLinha.esq, W, l);
	      uLinha = *(uLinha.dir);
	    }
	    else
	      uLinha = *(uLinha.esq);
      }
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_max].get_pd())){
	    l.push_back(uLinha.p_max);
      }
      if (ArvorePrioridadeyMSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_max(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pd())){
	      l.push_back(uLinha.esq->p_max);
	    }
      }
    }
    /*Neste ponto u eh ou uma folha ou uma quase folha.*/
    else {
      if (W.contem(vetorGlobal[u.p_max].get_pd())){
	    l.push_back(u.p_max);
      }
      if (ArvorePrioridadeyMSIndireto::isQuaseFolha(u)){
	    /*Neste ponto u eh quase folha. Então p_max(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[u.p_med].get_pd())){
	      l.push_back(u.esq->p_max);
	    }
      }
    }
  }
 
  static void PontosSubarvore (Segmento * vetorGlobal, noArvorePrioridadeyMSIndireto * v, Janela W, list <int> & l){
    if (v != NULL){
      if(vetorGlobal[v->p_max].get_pd().yMaior(W.getPontoInferiorEsquerdo()) || vetorGlobal[v->p_max].get_pd().Igual(W.getPontoInferiorEsquerdo())){
        l.push_back(v->p_max);
	    PontosSubarvore (vetorGlobal, v->esq, W, l);
	    PontosSubarvore (vetorGlobal, v->dir, W, l);
      }
    }
  }
  
  static void ImprimeArvorePrioridadeyMS (Segmento * vetorGlobal, noArvorePrioridadeyMSIndireto v){
    Ponto::ImprimePonto(vetorGlobal[v.p_max].get_pd());printf(" -> ");
    Segmento::ImprimeSegmento(vetorGlobal[v.p_max]);printf(" ");
    Ponto::ImprimePonto(vetorGlobal[v.p_med].get_pd());
    if (v.esq != NULL)
      ImprimeArvorePrioridadeyMS(vetorGlobal, *v.esq);
    if (v.dir != NULL)
      ImprimeArvorePrioridadeyMS(vetorGlobal, *v.dir);
  }

  static bool isFolha(noArvorePrioridadeyMSIndireto v){
    return (v.dir == NULL && v.esq == NULL);
  }

  static bool isQuaseFolha(noArvorePrioridadeyMSIndireto v){
    return (v.esq != NULL && v.dir == NULL);
  }
  
  static void inicializa (noArvorePrioridadeyMSIndireto * T){
    noArvorePrioridadeyMSIndireto * v = T;
    if (v != NULL){
      ArvorePrioridadeyMSIndireto::inicializa(v->esq);
      ArvorePrioridadeyMSIndireto::inicializa(v->dir);
      free(v);
    }
  }
  
  static size_t consumo_espaco (noArvorePrioridadeyMSIndireto * T){
  	size_t c = 0;
    noArvorePrioridadeyMSIndireto * v = T; 
    if (v != NULL){
      c = sizeof (v) + sizeof(v->p_med) + sizeof(v->p_max);
      c += ArvorePrioridadeyMSIndireto::consumo_espaco(v->esq);
      c += ArvorePrioridadeyMSIndireto::consumo_espaco(v->dir);
    }
    return c;
  }
};
