/*************************************************************************************************************************
  Seja P um conjunto de pontos. Seja W uma janela com w_1 = (x, -00) e w_2 = (x', y').
  Queremos encontrar os pontos de P que pertencem a W.
  Obs.: Uma arvore de busca com prioridade desta classe tem armazenado também segmentos.
 ************************************************************************************************************************/
struct noArvorePrioridadeymSIndireto {
  int p_min;
  int p_med;
  struct noArvorePrioridadeymSIndireto * dir;
  struct noArvorePrioridadeymSIndireto * esq;
};

class ArvorePrioridadeymSIndireto {
 public:
  /*****************************************************************************************************************
    Recebe um conjunto P de pontos representados por dois vetores P_x (x-ordenado) e P_y (y-ordenado) e um vetor de 
    segmentos S (ordenado de forma crescente (y-coordenada) por seus pontos extremos direitos) cujo pontos extremos 
    esquerdos são os pontos de P_x e P_y.
    Devolve uma arvore de busca com prioridade.
  ******************************************************************************************************************/
  static noArvorePrioridadeymSIndireto * ConstroiArvorePrioridadeymS (Segmento * vetorGlobal, int * Px, int * Py, int tamanho, char * ponto){
    noArvorePrioridadeymSIndireto * v;
    
    if (tamanho > 0){
      v = (noArvorePrioridadeymSIndireto *) malloc (sizeof(noArvorePrioridadeymSIndireto));
      int meio = (int) (((tamanho-1)/2.0)+0.5);
      int deslocamento = 0;
      int i, indexx, indexy_e, indexy_d;
      
      int * Py_esq = (int *) malloc (sizeof(int)* meio);
      int * Py_dir = (int *) malloc (sizeof(int)* (tamanho - 1 - meio));
      int * Px_esq = (int *) malloc (sizeof(int)* meio);
      int * Px_dir = (int *) malloc (sizeof(int)* (tamanho - 1 - meio));
      
      v->p_min = Py[0];
            
      for (i = 0, indexx = 0; i < meio + deslocamento; i++)
	    if (!(vetorGlobal[Px[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_min].qualPonto(ponto)))){
	      Px_esq[indexx] = Px[i];
	      indexx++;
	    }
	    else
	      deslocamento = 1;
      indexx = 0;
      while (i < tamanho){
	    if (!(vetorGlobal[Px[i]].qualPonto(ponto).Igual(vetorGlobal[v->p_min].qualPonto(ponto)))){
	      Px_dir[indexx] = Px[i];
	      indexx++;
	    }
	    i++;
      }
      
      v->p_med = Px[meio-1+deslocamento];
      
      for (i = 1, indexy_e = 0, indexy_d = 0; i < tamanho; i++){
	    if (vetorGlobal[Py[i]].qualPonto(ponto).xMenorIgual(vetorGlobal[v->p_med].qualPonto(ponto))){
	      Py_esq[indexy_e] = Py[i];
	      indexy_e++;
	    }
	    else{
	      Py_dir[indexy_d] = Py[i];
	      indexy_d++;
	    }
      }

      v->esq = ConstroiArvorePrioridadeymS (vetorGlobal, Px_esq, Py_esq, meio, ponto);
      v->dir = ConstroiArvorePrioridadeymS (vetorGlobal, Px_dir, Py_dir, tamanho - 1 - meio, ponto);
      free(Py_esq);
      free(Py_dir);
      free(Px_esq);
      free(Px_dir);
    }
    else
      v = NULL;
    return v;
  }
  
  static void ConsultaArvorePrioridadeymS (Segmento * vetorGlobal, noArvorePrioridadeymSIndireto v, Janela W, list <int> &l){
        
    noArvorePrioridadeymSIndireto u = v;
    noArvorePrioridadeymSIndireto uLinha;
    /*Pontos w1 e w2 de uma janela ilimitada no eixo x para -\infty.*/
    Ponto w1 = W.getPontoInferiorEsquerdo();
    Ponto w2 = W.getPontoSuperiorDireito();

    while (!(ArvorePrioridadeymSIndireto::isFolha(u)) && !(ArvorePrioridadeymSIndireto::isQuaseFolha(u)) 
	   && (w1.xMaior(vetorGlobal[u.p_med].get_pe()) || w2.xMenorIgual(vetorGlobal[u.p_med].get_pe()))){
            
      /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
      if (W.contem(vetorGlobal[u.p_min].get_pe())){
	    l.push_back(u.p_min);
      }
      
      if (w1.xMaior(vetorGlobal[u.p_med].get_pe()))
	    u = *(u.dir);
      else
	    u = *(u.esq);
    }
   

    if(!(ArvorePrioridadeymSIndireto::isFolha(u)) && !(ArvorePrioridadeymSIndireto::isQuaseFolha(u))){

      /*Verifica se u=v_div está em W.*/
      if (W.contem(vetorGlobal[u.p_min].get_pe())){
	    l.push_back(u.p_min);
      }

      /*Ande para a esquerda de u.*/
      uLinha = *(u.esq);
      while (!(ArvorePrioridadeymSIndireto::isFolha(uLinha)) && !(ArvorePrioridadeymSIndireto::isQuaseFolha(uLinha))){
      
        /*Neste ponto verificamos se o ponto armazenado no no' uLinha esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	      l.push_back(uLinha.p_min);
	    }
	
	    if (w1.xMenorIgual(vetorGlobal[uLinha.p_med].get_pe())){
	      ArvorePrioridadeymSIndireto::PontosSubarvore(vetorGlobal, uLinha.dir, W, l);
	      uLinha = *(uLinha.esq);
	    }
	    else
	      uLinha = *(uLinha.dir);
      }
      
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	    l.push_back(uLinha.p_min);
      }
      if (ArvorePrioridadeymSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_min(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pe())){
	      l.push_back(uLinha.esq->p_min);
	    }
      }
      
      /*Ande para a direita de u.*/
      uLinha = *(u.dir);
      while (!(ArvorePrioridadeymSIndireto::isFolha(uLinha)) && !(ArvorePrioridadeymSIndireto::isQuaseFolha(uLinha))){
	
	    /*Neste ponto verificamos se o ponto armazenado no no' u esta' em W.*/
	    if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	      l.push_back(uLinha.p_min);
	    }
 
    	if (w2.xMaior(vetorGlobal[uLinha.p_med].get_pe())){
	      ArvorePrioridadeymSIndireto::PontosSubarvore(vetorGlobal, uLinha.esq, W, l);
	      uLinha = *(uLinha.dir);
	    } 
		else
	      uLinha = *(uLinha.esq);
      }
      /*Neste ponto uLinha eh quase folha ou folha.*/
      if (W.contem(vetorGlobal[uLinha.p_min].get_pe())){
	    l.push_back(uLinha.p_min);
      }
      if (ArvorePrioridadeymSIndireto::isQuaseFolha(uLinha)){
	    /*Neste ponto uLinha eh quase folha. Então p_min(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[uLinha.p_med].get_pe())){
	      l.push_back(uLinha.esq->p_min);
	    }
      }
    }
    /*Neste ponto u eh ou uma folha ou uma quase folha.*/
    else {
      if (W.contem(vetorGlobal[u.p_min].get_pe())){
	    l.push_back(u.p_min);
      }
      if (ArvorePrioridadeymSIndireto::isQuaseFolha(u)){
	    /*Neste ponto u eh quase folha. Então p_min(abaixo(u))=p_med(u).*/
	    if (W.contem(vetorGlobal[u.p_med].get_pe())){
	      l.push_back(u.esq->p_min);
    	} 
      }
    }
  }
  
  static void PontosSubarvore (Segmento * vetorGlobal, noArvorePrioridadeymSIndireto * v, Janela W, list <int> & l){
    if (v != NULL){
      if(vetorGlobal[v->p_min].get_pe().yMenorIgual(W.getPontoSuperiorDireito())){
        //if(v->p_min.getX()<=(W.getPontoSuperiorDireito().getX())){
	    /*l.push_back(v->p_min);*/
	    l.push_back(v->p_min);
	    PontosSubarvore (vetorGlobal, v->esq, W, l);
	    PontosSubarvore (vetorGlobal, v->dir, W, l);
      }
    }
  }

  static void ImprimeArvorePrioridadeymS (Segmento * vetorGlobal, noArvorePrioridadeymSIndireto v){
    Ponto::ImprimePonto(vetorGlobal[v.p_min].get_pe());printf(" -> ");
    Segmento::ImprimeSegmento(vetorGlobal[v.p_min]);printf(" ");
    Ponto::ImprimePonto(vetorGlobal[v.p_med].get_pe());
    if (v.esq != NULL)
      ImprimeArvorePrioridadeymS(vetorGlobal, *v.esq);
    if (v.dir != NULL)
      ImprimeArvorePrioridadeymS(vetorGlobal, *v.dir);
  }

  static bool isFolha(noArvorePrioridadeymSIndireto v){
    return (v.dir == NULL && v.esq == NULL);
  }

  static bool isQuaseFolha(noArvorePrioridadeymSIndireto v){
    return (v.esq != NULL && v.dir == NULL);
  }

  static void inicializa (noArvorePrioridadeymSIndireto * T){
    noArvorePrioridadeymSIndireto * v = T;
    if (v != NULL){
      ArvorePrioridadeymSIndireto::inicializa(v->esq);
      ArvorePrioridadeymSIndireto::inicializa(v->dir);
      free(v);
    }
  }
  
  static size_t consumo_espaco (noArvorePrioridadeymSIndireto * T){
  	size_t c = 0;
    noArvorePrioridadeymSIndireto * v = T; 
    if (v != NULL){
      c = sizeof(v) + sizeof(v->p_med) + sizeof(v->p_min);
      c += ArvorePrioridadeymSIndireto::consumo_espaco(v->esq);
      c += ArvorePrioridadeymSIndireto::consumo_espaco(v->dir);
    }
    return c;
  }
};
